# Repository path:    $HeadURL: https://90.145.156.222/DMS-Houtribdijk/FEWS/Modules/Waterlevels/PressureCorrection.py $
# Last committed:     $Revision: 420908 $; $Author: vvuik $; $Date: 2019-11-15 17:10:26 +0100 (vr, 15 nov 2019) $
#
# -*- coding: utf-8 -*-
"""
Created on Thu Jul 11 11:33:58 2019

@author: hove
"""

from waterlevels import ReadMetaDataAdjusted
from waterlevels import Prcorr
from waterlevels import WriteNcdfFile
from waterlevels import FewsLogger
import os.path
from scipy import signal
from shutil import copy
import getopt
import netCDF4
import numpy as np
import pandas as pd
import sys
import datetime
import itertools

class NoDataException(Exception):
    """Base class for other exceptions"""
    pass


def get_location(data_set, log):
    location = ""
    loc_array = data_set.variables['station_id'][:]
    for i in range(loc_array.count(axis=None)):
        location = location + loc_array[0][i].decode("utf-8")
    log.info("Location in netcdf is {}".format(location))
    return location


def apply_fft(df, H_water_down, Z_bedlevel, bNum, location):
    # Parameters
    rho = 1000
    g = 9.81
    Fs = 4  # sampling frequency of sensor
    maxE = 0.01  # maximum fraction of empty values in bin
    eMax = np.ceil(maxE * bNum)  # maximum number of empty values in bin

    # Check for outliers
    mu = np.nanmean(df['D.water.up'])  # mean of bin
    sigma = np.nanstd(df['D.water.up'])  # standard deviation of bin
    limit = mu + 4 * sigma
    # replace values with np.nan if outside limit (= outlier)
    df['D.water.up'] = [i if i <= limit else np.nan for i in df['D.water.up'].values]
    # make column with D.water.up value of previous time step
    df['D.previous'] = [df.iloc[i - 1]['D.water.up'] if i > 0 else df.iloc[i]['D.water.up'] for i in range(len(df))]
    # if difference in D.water.up is larger than 4*sigma, replace D.water.up with np.nan
    df['D.water.up'] = [df.iloc[i]['D.water.up'] if abs(df.iloc[i]['D.water.up'] - df.iloc[i]['D.previous']) < (4 * sigma) else np.nan for i in range(len(df))]

    eNan = df['D.water.up'].isnull().sum()  # number of nan values in sample

    if (eNan < eMax) and (len(df) > 5*60*4):
        
        df['D.water.up'].interpolate(method='time', limit_direction='both', inplace=True)
        
        df['H.water.up'] = df['D.water.up'] * 100 / rho / g  # m
        df['H.water.total'] = df['H.water.up'] + H_water_down  # m
        # Average height of total watercolumn
        Hmean = np.nanmean(df['H.water.total'])  # m

        if location=='BUOY1':
            waterlevelfft =  df['H.water.total'].values + Z_bedlevel
        else:
            # Detrend = remove linear trend along axis from data
            zsd = signal.detrend(df['H.water.up'].values)
            zsc = Prcorr.prcorr(zsd, Hmean, Fs, H_water_down, nsample=2 ** 8, fcorrmin=0.05, fcorrmax=1.00, maxcorr=5)
            waterlevelfft = Z_bedlevel + Hmean + zsc
        return waterlevelfft
    else:
        print('gives back NaN')
        return np.full(bNum, np.nan)
    
def pressure_correction(file_metadata, instrument, logfile):

    parid = 'D_0_water'

    input_file = r".\input\input.nc"
    output_file = r".\output\output.nc"

    log = FewsLogger.Logger(logfile)
    try:
        # check file exists
        if not os.path.isfile(input_file):
            log.error("Input file not found ({})".format(input_file))
            raise FileNotFoundError

        # Read netcdf
        log.info("Reading input file {}".format(input_file))
        dataSet = netCDF4.Dataset(input_file)
        time_min = dataSet.variables['time'][:].data
        time_sec = time_min * 60
        index = pd.to_datetime(time_sec, unit='s')
        if len(index) == 0:
            raise NoDataException
        else:
            log.debug("The number of time steps in {} is {}".format(input_file, len(index)))

        # Get location from netcdf
        location = get_location(dataSet, log)

        # Dataframe with index and waterlevels
        data = dataSet.variables[parid][:].data
        data[data<0] = np.NaN # change -999 into NaN (and negative pressure is impossible)
        df_input = pd.DataFrame(index=index, data={'D.water.up': data[: ,0]})

        startTime = index[0]
        year = startTime.year
        month = startTime.month
        day = startTime.day
        date_range = pd.date_range(start=datetime.date(year, month, day), end=datetime.date(year, month, day) + datetime.timedelta(days=1), freq='15min')

        # Read metadata
        # Get bed level and z-location of sensor
        log.info("Reading metadata from location {}".format(file_metadata))
        read_meta_data = ReadMetaDataAdjusted.ReadMetaDataAdjusted(file_metadata, [startTime], log, location)
        meta_data = read_meta_data.get_meta_data()

        if instrument=='ADV':
            # ADV: hoogte meetvolume voor snelheden in metadatasheet
            dzpres = 0.368 # afstand pressure sensor boven meetvolume ADV
        elif instrument=='AQD':
            # AQD: hoogte transducers en druksensor (=identiek) in metadatasheet
            dzpres = 0 # pressure sensor naast transducers bij AQD
            
        Zmetadata = meta_data[location]['{}_NAP_Z'.format(instrument)] # hoogtemeting in metadatasheet (m+NAP)
        Zdruksensor = Zmetadata+dzpres # hoogte druksensor (m+NAP)
        Z_bedlevel = meta_data[location]['bodem_NAP_Z'] # bodemhoogte (m+NAP)
        H_water_down = Zdruksensor - Z_bedlevel # waterdiepte onder druksensor (m)

        log.info("Starting computations")
        wave_data = []
        for i in range(len(date_range)-1):

            # define a bin range and reindex
            bin_range = pd.date_range(start = date_range[i], end = date_range[i+1], freq = '250ms', closed = 'left')
            df_15min = df_input.reindex(bin_range)
            
            try:
                bin_start = df_15min[df_15min['D.water.up'].notnull()].index[0] # timestamp of first non NaN-value
                bin_end = df_15min[df_15min['D.water.up'].notnull()].index[-1] # timestamp of last non NaN-value
                df_temp = df_15min.loc[bin_start:bin_end,:]
                Zcorrected = apply_fft(df_temp, H_water_down, Z_bedlevel, len(df_temp), location) # compute waterlevel
                Zstart = np.full(len(pd.date_range(start = date_range[i], end = bin_start, freq='250ms', closed = 'left')), np.nan) 
                Zend = np.full(len(pd.date_range(start = bin_end, end = date_range[i+1], freq='250ms', closed = 'left'))-1, np.nan)
                Z = np.concatenate((Zstart, Zcorrected, Zend), axis=None)
                wave_data.append(Z)
            except:
                Z = np.full(len(df_15min), np.nan)
                wave_data.append(Z)

        Z_water = list(itertools.chain(*wave_data))
        date_data = pd.date_range(start=datetime.date(year, month, day), end=datetime.date(year, month, day) + datetime.timedelta(days=1), freq='250ms', closed='left')
        df = pd.DataFrame(index = date_data , data = {'Z.water.total.corrected': Z_water})

        # write data to netcdf
        log.info("Writing output file {}".format(output_file))
        WriteNcdfFile.write(output_file, df, instrument, location, parid)
    except NoDataException:
        log.warning("The number of time steps in {} is {}. Skipping this location".format(input_file, len(index)))
        # copy input file to output folder to avoid a crash at the import
        copy(input_file, output_file)
    finally:
        log.close()

def main(argv):
    # input argument checking
    try:
        opts, args = getopt.getopt(argv, "hm:i:l:", ["help", "metadata_file=", "instrument=", "logfile="])
        if len(argv) < 5:
            print("PressureCorrection.py called without the corrected number of arguments")
            print("use: PressureCorrection.py -m <metadata_file> -i [ADV|AQD] -l logfile")
            raise getopt.GetoptError("Arguments expected")
    except getopt.GetoptError:
        print("PressureCorrection.py -m <metadata_file> -i [ADV|AQD] -l logfile")
        print("PressureCorrection.py called without the corrected number of arguments")
        print("use: PressureCorrection.py -m <metadata_file> -i [ADV|AQD] -l logfile")
    else:
        for opt, arg in opts:
            if opt in ("-h", "help"):
                print("PressureCorrection.py -m <metadata_file> -i [ADV|AQD] -l logfile")
                sys.exit()
            elif opt in ("-m", "--metadata_file"):
                file_metadata = arg
            elif opt in ("-i", "--instrument"):
                instrument = arg
            elif opt in ("-l", "--logfile"):
                logfile = arg

        # print ("PressureCorrection.py -m {} -i {}".format(file_metadata, instrument))
        pressure_correction(file_metadata, instrument, logfile)


if __name__ == "__main__":
    main(sys.argv[1:])


